"""Analyze image savig it as numpy file."""

import argparse

import cv2
import numpy as np
from correct_image import Formatter
from loader import get_loader, Loader

from libs.npyplotter.npyplotter.plot_npy import picshow

FOLDER_PATH = '/home/joseph/Documents/Thesis/Dataset/config'


def main(
    config_path: str,
    video_identifier: str,
    entire_frame=False,
    undistort=True,
    roi=True,
    get_frame=True,
    plot=False
) -> None:
    """Save the first image as numpy file."""
    loader: Loader = get_loader(config_path, video_identifier)
    formatter = Formatter(config_path, video_identifier)
    image: np.ndarray = loader.read()
    if get_frame:
        cv2.imwrite('image.png', image)
    if entire_frame:
        formatter.show_entire_image()
    if undistort:
        image = formatter.apply_distortion_correction(image)
    if roi:
        image = formatter.apply_roi_extraction(image)
    if get_frame:
        cv2.imwrite('image.png', image)
    np.save('tmp.npy', image)
    if plot:
        picshow([image])


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'statio_name',
        help='Name of the station to be analyzed')
    parser.add_argument(
        'video_identifier',
        help='Index of the video of the json config file')
    parser.add_argument(
        '-f',
        '--frame',
        action='store_true',
        help='Plot entire frame or not')
    parser.add_argument(
        '-u',
        '--undistort',
        action='store_true',
        help='Format image using distortion correction')
    parser.add_argument(
        '-g',
        '--getframe',
        action='store_true',
        help='Get first frame')
    parser.add_argument(
        '-r',
        '--roi',
        action='store_true',
        help='Format image using selecting only roi area')
    parser.add_argument(
        '-P',
        '--plot',
        action='store_true',
        help='Plot output image')
    args = parser.parse_args()
    CONFIG_PATH = f'{FOLDER_PATH}/{args.statio_name}.json'
    main(
        config_path=CONFIG_PATH,
        video_identifier=args.video_identifier,
        entire_frame=args.frame,
        undistort=args.undistort,
        roi=args.roi,
        get_frame=args.getframe,
        plot=args.plot
    )



"""Optical Tracking Image Velocimetry."""

import argparse
import math
import random
from typing import NamedTuple, Optional

import cv2
import numpy as np
from numpy.typing import NDArray
from pydantic import BaseModel

from awive.algorithms.image_velocimetry import ImageVelocimetry
from awive.config import Config, ConfigOtv, ConfigRoi
from awive.correct_image import Formatter
from awive.loader import Loader, get_loader


FOLDER_PATH = "examples/datasets"


class Stats(NamedTuple):
    """Stats of the algorithm."""

    avg: float = 0
    max_: float = 0
    min_: float = 0
    std_dev: float = 0
    count: int = 0


def get_magnitude(kp1, kp2):
    """Get the distance between two keypoints."""
    # return math.sqrt((kp2.pt[1] - kp1.pt[1])**2 + \
    # (kp2.pt[0] - kp1.pt[0]) ** 2 )
    return abs(kp2.pt[0] - kp1.pt[0])


def get_angle(kp1, kp2) -> float:
    """Get angle between two key points."""
    return math.atan2(
        kp2.pt[1] - kp1.pt[1],
        kp2.pt[0] - kp1.pt[0]
    ) * 180 / math.pi


def _get_velocity(kp1, kp2, real_distance_pixel, time, fps):
    if time == 0:
        return 0
    return get_magnitude(kp1, kp2) * real_distance_pixel * fps / time


def reject_outliers(data, m=2.):
    """Reject outliers of data."""
    d = np.abs(data - np.median(data))
    mdev = np.median(d)
    s = d / mdev if mdev else 0.
    return data[s < m]


def compute_stats(velocity, hist=False) -> Stats:
    """Compute stats of list of velicities."""
    v = np.array(sum(velocity, []))
    v = reject_outliers(v)
    count = len(v)
    if count == 0:
        return Stats()
    avg: float = v.mean()
    max_: float = v.max()
    min_: float = v.min()
    std_dev: float = np.std(v)  # type: ignore

    if hist:
        import matplotlib.pyplot as plt
        plt.hist(v.astype(int))
        plt.ylabel('Probability')
        plt.xlabel('Data')
        plt.show()

    return Stats(avg, max_, min_, std_dev, count)


class LKParams(BaseModel):
    """Lucas-Kanade parameters."""

    winSize: tuple[int, int]
    maxLevel: int
    criteria: tuple[int, int, float]
    flags: int
    minEigThreshold: float


class Otv(ImageVelocimetry):
    """Optical Tracking Image Velocimetry."""

    def __init__(self, config: Config, debug: bool) -> None:
        """Initialize OTV."""
        self.conf: ConfigOtv = config.otv
        self._debug: bool = debug
        self._partial_max_angle: float = self.conf.partial_max_angle
        self._partial_min_angle: float = self.conf.partial_min_angle
        self._final_max_angle: float = self.conf.final_max_angle
        self._final_min_angle: float = self.conf.final_min_angle
        self._final_min_distance: float = self.conf.final_min_distance
        self._max_features: int = self.conf.max_features
        self._radius: int = self.conf.lk.radius
        self._max_level: int = self.conf.lk.max_level
        self._step: int = self.conf.region_step
        self._resolution: int = self.conf.resolution
        self._pixel_to_real: float = self.conf.pixel_to_real

        roi: ConfigRoi = config.preprocessing.roi
        self._width: int = roi.w2 - roi.w1
        self._height: int = roi.h2 - roi.h1
        self._regions: list[int] = self.conf.lines

        # Load mask if exists
        self._mask: Optional[NDArray] = None
        if self.conf.mask_path:
            self._mask = cv2.imread(self.conf.mask_path, 0) > 1  # type: ignore
            self._mask = cv2.resize(
                self._mask.astype(np.uint8),
                (self._height, self._width),
                cv2.INTER_NEAREST  # type: ignore
            )

        # Load Lucas Kanade parameters
        self.lk_params: LKParams = LKParams(
            winSize=(self._radius, self._radius),
            maxLevel=self._max_level,
            criteria=(
                cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT,
                self.conf.lk.max_count,
                self.conf.lk.epsilon
            ),
            flags=cv2.OPTFLOW_LK_GET_MIN_EIGENVALS,
            minEigThreshold=self.conf.lk.min_eigen_threshold
        )

        # Load firt images
        # self.prev_gray = prev_gray

    def _partial_filtering(self, kp1, kp2, max_distance):
        # only to limit the research window
        magnitude = get_magnitude(kp1, kp2)
        if magnitude > max_distance:
            return False
        angle = get_angle(kp1, kp2)
        if angle < 0:
            angle = angle + 360
        if angle == 0 or angle == 360:
            return True
        if self._partial_min_angle <= angle <= self._partial_max_angle:
            return True
        return False

    def _final_filtering(self, kp1, kp2):
        """Final filter of keypoints."""
        magnitude = get_magnitude(kp1, kp2)
        if magnitude < self._final_min_distance:
            return False
        angle = get_angle(kp1, kp2)
        if angle < 0:
            angle = angle + 360
        if angle == 0 or angle == 360:
            return True
        if self._final_min_angle <= angle <= self._final_max_angle:
            return True
        return False

    def _apply_mask(self, image):
        if self._mask is not None:
            image = image * self._mask
        return image

    def _init_subregion_list(self, dimension, width):
        ret = []
        n_regions = math.ceil(width / self._step)
        for _ in range(n_regions):
            # TODO: This is so inneficient
            if dimension == 1:
                ret.append(0)
            elif dimension == 2:
                ret.append([])
        return ret

    def run(self, loader: Loader, formatter: Formatter, show_video=False):
        """Execute OTV and get velocimetry."""
        # initialze parametrers
        detector = cv2.FastFeatureDetector_create()
        previous_frame = None
        keypts_curr: list[cv2.KeyPoint] = []  # current keypoints
        keypoints_start: list[cv2.KeyPoint] = []
        time: list[int] = []  # list of indices of Keypoints
        keypts_pred: list[cv2.KeyPoint] = []  # predicted keypoints
        masks: list[NDArray[np.uint8]] = []

        valid: list[list[bool]] = []  # Indices of valid Keypoints
        velocity_mem: list[list[float]] = []
        keypoints_mem_current: list[cv2.KeyPoint] = []
        keypoints_mem_predicted: list[cv2.KeyPoint] = []
        velocity: list[list[float]] = []
        angle: list[list[float]] = []
        distance: list[list[float]] = []
        path: list[list[int]] = []
        traj_map = np.zeros((1200, 900))

        # update width and height if needed
        if loader.image_shape[0] < self._width:
            self._width = loader.image_shape[0]
        if loader.image_shape[1] < self._height:
            self._height = loader.image_shape[1]

        # This are not used. but idk why :c
        # subregion_velocity = self._init_subregion_list(2, self._width)
        # subregion_trajectories = self._init_subregion_list(1, self._width)

        regions: list[list[float]] = [[] for _ in range(len(self._regions))]

        # Initialization
        for i in range(loader.total_frames):
            valid.append([])
            velocity_mem.append([])
            velocity.append([])
            angle.append([])
            distance.append([])
            path.append([])

        while loader.has_images():
            # get current frame
            current_frame = loader.read()
            current_frame = formatter.apply_distortion_correction(
                current_frame)
            current_frame = formatter.apply_roi_extraction(current_frame)
            # current_frame = self._apply_mask(current_frame)

            # get features as a list of KeyPoints
            keypoints: list[cv2.KeyPoint] = detector.detect(
                current_frame,
                None
            )
            random.shuffle(keypoints)

            # update lot of lists
            if previous_frame is None:
                for i, keypoint in enumerate(keypoints):
                    if len(keypts_curr) < self._max_features:
                        keypts_curr.append(keypoint)
                        keypoints_start.append(keypoint)
                        time.append(loader.index)
                        valid[loader.index].append(False)
                        velocity_mem[loader.index].append(0)
                        path[loader.index].append(i)
            else:
                for i, keypoint in enumerate(reversed(keypoints)):
                    if len(keypts_curr) < self._max_features:
                        keypts_curr.append(keypoint)
                        keypoints_start.append(keypoint)
                        time.append(loader.index)
                        valid[loader.index].append(False)
                        velocity_mem[loader.index].append(0)

            if self._debug >= 1:
                print('Analyzing frame:', loader.index)
            if previous_frame is not None:
                pts1 = cv2.KeyPoint_convert(keypts_curr)
                pts2, st, _ = cv2.calcOpticalFlowPyrLK(
                    previous_frame,
                    current_frame,
                    pts1,
                    None,
                    **self.lk_params.dict()
                )

                # add predicted by Lucas-Kanade new keypoints
                keypts_pred.clear()
                for pt2 in pts2:
                    keypts_pred.append(cv2.KeyPoint(
                        pt2[0],
                        pt2[1],
                        1.0
                    ))

                max_distance = self._max_level * (2 * self._radius + 1)
                max_distance /= self._resolution

                k = 0

                for i, keypoint in enumerate(keypts_curr):
                    partial_filter = self._partial_filtering(
                        keypoint,
                        keypts_pred[i],
                        max_distance
                    )
                    # check if the trajectory finished or the vector is invalid
                    if not (st[i] and partial_filter):
                        final_filter = self._final_filtering(
                            keypoints_start[i],
                            keypts_curr[i]
                        )
                        # check if it is a valid trajectory
                        if final_filter:
                            velocity_i = _get_velocity(
                                keypoints_start[i],
                                keypts_curr[i],
                                self._pixel_to_real / self._resolution,
                                loader.index - time[i],
                                loader.fps
                            )
                            angle_i: float = get_angle(
                                keypoints_start[i],
                                keypts_curr[i]
                            )

                            xx0 = int(keypoints_start[i].pt[1])
                            yy0 = int(keypoints_start[i].pt[0])
                            traj_map[xx0][yy0] += 100
                            # sub-region computation
                            # module_start = int(keypoints_start[i].pt[1] /
                            #         self._step)
                            # module_current = int(keypoints_current[i].pt[1] /
                            #         self._step)
                            # if module_start == module_current:
                            # subregion_velocity[module_start].append(velocity_i)
                            # subregion_trajectories[module_start] += 1

                            for r_idx, region in enumerate(self._regions):
                                if abs(xx0 - region) < 15:
                                    regions[r_idx].append(velocity_i)

                            # update storage
                            pos = i
                            j = loader.index - 1
                            while j >= time[i]:
                                valid[j][pos] = True
                                velocity_mem[j][pos] = velocity_i
                                pos = path[j][pos]
                                j -= 1

                            velocity[loader.index].append(velocity_i)
                            angle[loader.index].append(angle_i)
                            distance[loader.index].append(
                                velocity_i *
                                (loader.index - time[i]) /
                                loader.fps
                            )

                        continue

                    # Add new displacement vector
                    keypts_curr[k] = keypts_curr[i]
                    keypoints_start[k] = keypoints_start[i]
                    keypts_pred[k] = keypts_pred[i]
                    path[loader.index].append(i)
                    velocity_mem[loader.index].append(0)
                    valid[loader.index].append(False)
                    time[k] = time[i]
                    k += 1

                # Only keep until the kth keypoint in order to filter invalid
                # vectors
                keypts_curr = keypts_curr[:k]
                keypoints_start = keypoints_start[:k]
                keypts_pred = keypts_pred[:k]
                time = time[:k]

            if self._debug >= 1:
                print('number of trajectories:', len(keypts_curr))

            if show_video:
                if previous_frame is not None:
                    color_frame = cv2.cvtColor(
                        current_frame, cv2.COLOR_GRAY2RGB)
                    output = draw_vectors(
                        color_frame,
                        keypts_pred,
                        keypts_curr,
                        masks
                    )
                    cv2.imshow("sparse optical flow", output)
                    if cv2.waitKey(10) & 0xFF == ord('q'):
                        break
            previous_frame = current_frame.copy()
            keypoints_mem_current.append(keypts_curr)
            keypoints_mem_predicted.append(keypts_pred)

            # TODO: I guess the swap is not needed such as in the next
            # iteration the keypoints_predicted will be cleaned
            if len(keypts_pred) != 0:
                keypts_pred, keypts_curr = keypts_curr, keypts_pred
        np.save('traj.npy', traj_map)

        loader.end()
        cv2.destroyAllWindows()
        avg, max_, min_, std_dev, count = compute_stats(velocity, show_video)

        if self._debug >= 1:
            print('avg:', round(avg, 4))
            print('max:', round(max_, 4))
            print('min:', round(min_, 4))
            print('std_dev:', round(std_dev, 2))
            print('count:', count)

        out_json: dict = {}
        for i, sv in enumerate(regions):
            out_json[str(i)] = {}
            t = np.array(sv)
            t = t[t != 0]
            if len(t) != 0:
                t = reject_outliers(t)
                m = t.mean()
            else:
                m = 0
            out_json[str(i)]['velocity'] = m
            out_json[str(i)]['count'] = len(t)
        return out_json


def draw_vectors(image, new_list, old_list, masks):
    """Draw vectors of velocity and return the output and update mask"""
    if len(image.shape) == 3:
        color = (0, 255, 0)
        thick = 1
    else:
        color = 255
        thick = 1

    # create new mask
    mask = np.zeros(image.shape, dtype=np.uint8)
    for new, old in zip(new_list, old_list):
        # TODO: Perhaps there is a more efficient way to transform this to a
        # numpy array
        new_pt = np.array([int(new.pt[0]), int(new.pt[1])])
        old_pt = np.array([int(old.pt[0]), int(old.pt[1])])
        mask = cv2.line(mask, new_pt.ravel(), old_pt.ravel(), color, thick)

    # update masks list
    masks.append(mask)
    if len(masks) < 3:
        return np.zeros(image.shape)
    if len(masks) > 3:
        masks.pop(0)

    # generate image with mask
    total_mask = np.zeros(mask.shape, dtype=np.uint8)
    for mask_ in masks:
        total_mask = cv2.add(total_mask, mask_)
    output = cv2.add(image, total_mask)
    return output


def main(
    config_path: str,
    video_identifier: str,
    show_video: bool = False,
    debug: bool = False
):
    """Execute a basic example of OTV."""
    loader: Loader = get_loader(config_path, video_identifier)
    formatter: Formatter = Formatter(config_path, video_identifier)

    # loader.has_images()
    # image = loader.read()
    # prev_gray = formatter.apply_distortion_correction(image)
    # prev_gray = formatter.apply_roi_extraction(prev_gray)

    config = Config.from_json(config_path, video_identifier)
    otv = Otv(config, debug)
    return otv.run(loader, formatter, show_video)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "statio_name",
        help="Name of the station to be analyzed"
    )
    parser.add_argument(
        "video_identifier",
        help="Index of the video of the json config file"
    )
    parser.add_argument(
        '-p',
        '--path',
        help='Path to the config folder',
        type=str,
        default=FOLDER_PATH
    )
    parser.add_argument(
        '-d',
        '--debug',
        help='Activate debug mode',
        type=int,
        default=0
    )
    parser.add_argument(
        '-v',
        '--video',
        action='store_true',
        help='Play video while processing'
    )
    args = parser.parse_args()
    print(
        main(
            config_path=f'{args.path}/{args.statio_name}/config.json',
            video_identifier=args.video_identifier,
            show_video=args.video,
            debug=args.debug
        )
    )



"""Convert to video."""
import cv2 as cv
import numpy as np


# Define the codec and create VideoWriter object
xd = (421, 310)
fourcc = cv.VideoWriter_fourcc(*'XVID')
out = cv.VideoWriter('output.avi', fourcc, 20.0, xd)
for i in range(100):
    frame = np.load(f'images/im_{i:04d}.npy')
    frame = cv.resize(frame, xd, interpolation=cv.INTER_AREA)
    frame = cv.cvtColor(frame, cv.COLOR_GRAY2RGB)
    # write the flipped frame
    out.write(frame)
    cv.imshow('frame', frame)
    if cv.waitKey(1) == ord('q'):
        break
# Release everything if job is finished
out.release()
cv.destroyAllWindows()



"""Play  a video."""
import argparse
import json

import cv2
import numpy as np

from correct_image import Formatter
from loader import Loader, get_loader


FOLDER_PATH = '/home/joseph/Documents/Thesis/Dataset/config'
RESIZE_RATIO = 5


def play(
    loader: Loader,
    formatter: Formatter,
    undistort=True,
    roi=True,
    time_delay=1,
    resize=False,
    wlcrop=None,
    blur=True
) -> None:
    """Plays a video."""
    i: int = 0

    while loader.has_images():
        image = loader.read()
        if undistort:
            image = formatter.apply_distortion_correction(image)
        if roi:
            image = formatter.apply_roi_extraction(image)
        elif wlcrop is not None:
            image = image[wlcrop[0], wlcrop[1]]
        if blur:
            image = cv2.medianBlur(image, 5)
        if resize:
            lil_im = cv2.resize(image, (1000, 1000))
        else:
            lil_im = image
        cv2.imshow('Video', lil_im)
        np.save(f'images/im_{i:04}.npy', lil_im)
        if cv2.waitKey(time_delay) & 0xFF == ord('q'):
            print('Finished by key \'q\'')
            break
        i += 1
    cv2.destroyAllWindows()


def main(
    config_path: str,
    video_identifier: str,
    undistort=True,
    roi=True,
    time_delay=1,
    resize=True,
    wlcrop=True,
    blur=True
) -> None:
    """Read configurations and play video."""
    loader = get_loader(config_path, video_identifier)
    formatter = Formatter(config_path, video_identifier)
    if wlcrop:
        with open(config_path) as json_file:
            config = json.load(json_file)[video_identifier]['water_level']
        roi2 = config['roi']
        wr0 = slice(roi2[0][0], roi2[1][0])
        wr1 = slice(roi2[0][1], roi2[1][1])
        crop = (wr0, wr1)
    else:
        crop = None
    play(loader, formatter, undistort, roi, time_delay, resize, crop, blur)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "statio_name",
        help="Name of the station to be analyzed"
    )
    parser.add_argument(
        "video_identifier",
        help="Index of the video of the json config file"
    )
    parser.add_argument(
        '-u',
        '--undistort',
        action='store_true',
        help='Format image using distortion correction'
    )
    parser.add_argument(
        '-r',
        '--roi',
        action='store_true',
        help='Format image using selecting only roi area'
    )
    parser.add_argument(
        '-c',
        '--wlcrop',
        action='store_true',
        help='Water level crop'
    )
    parser.add_argument(
        '-b',
        '--blur',
        action='store_true',
        help='Blur image'
    )
    parser.add_argument(
        '-z',
        '--resize',
        action='store_true',
        help='Resizer image to 1000x1000'
    )
    parser.add_argument(
        '-t',
        '--time',
        default=1,
        type=int,
        help='Time delay between each frame (ms)'
    )
    args = parser.parse_args()
    main(
        config_path=f'{FOLDER_PATH}/{args.statio_name}.json',
        video_identifier=args.video_identifier,
        undistort=args.undistort,
        roi=args.roi,
        time_delay=args.time,
        resize=args.resize,
        wlcrop=args.wlcrop,
        blur=args.blur,
    )



"""Correct distortion of videos.

This module contains classes and functions needed to correct distortion of
videos, either intrinsic or extring to the camera. Also it saves the corrected
frames in defined directory path.

"""

import argparse
import json
import time
from typing import Any

import cv2
import numpy as np
from numpy.typing import NDArray

import awive.imageprep as ip
from awive.loader import get_loader

FOLDER_PATH = '/home/joseph/Documents/Thesis/Dataset/config'


class Formatter:
    """Format frames in order to be used by image processing methods."""

    def __init__(self, config_path: str, video_identifier: str) -> None:
        """Initialize Formatter object."""
        # read configuration file
        with open(config_path) as json_file:
            self._config = json.load(json_file)[video_identifier]

        sample_image = self._get_sample_image(config_path, video_identifier)
        self._shape = (sample_image.shape[0], sample_image.shape[1])
        if self._config["dataset"]['gcp']['apply']:
            self._or_params = self._get_orthorectification_params(sample_image)
        else:
            self._or_params = None

        self._rotation_angle = self._config["preprocessing"]['rotate_image']
        self._rotation_matrix = self._get_rotation_matrix()

        w_slice = slice(self._config["preprocessing"]['roi']['w1'],
                        self._config["preprocessing"]['roi']['w2'])
        h_slice = slice(self._config["preprocessing"]['roi']['h1'],
                        self._config["preprocessing"]['roi']['h2'])
        self._slice = (w_slice, h_slice)
        w_slice = slice(self._config["preprocessing"]['pre_roi']['w1'],
                        self._config["preprocessing"]['pre_roi']['w2'])
        h_slice = slice(self._config["preprocessing"]['pre_roi']['h1'],
                        self._config["preprocessing"]['pre_roi']['h2'])
        self._pre_slice = (w_slice, h_slice)

    def _get_orthorectification_params(
        self,
        sample_image: np.ndarray,
        reduce=None
    ) -> tuple[Any, NDArray]:
        x = self._config['gcp']['pixels']
        df_from = list(map(list, zip(*[(v) for k, v in x.items()])))
        if reduce is not None:
            for i, _ in enumerate(df_from):
                df_from[i][0] = df_from[i][0]-reduce[0]
                df_from[i][1] = df_from[i][1]-reduce[1]

        x = self._config["dataset"]['gcp']['meters']
        df_to = list(map(list, zip(*[(v) for k, v in x.items()])))
        if self._config["preprocessing"]['image_correction']['apply']:
            corr_img = ip.lens_corr(
                    sample_image,
                    k1=self._config["preprocessing"]['image_correction']['k1'],
                    c=self._config["preprocessing"]['image_correction']['c'],
                    f=self._config["preprocessing"]['image_correction']['f']
                    )
        else:
            corr_img = sample_image
        M, C, __ = ip.orthorect_param(corr_img,
                                      df_from,
                                      df_to,
                                      PPM=self._config["dataset"]['PPM'],
                                      lonlat=False)
        return (M, C)

    def _get_rotation_matrix(self):
        """Rotate matrix.

        based on:
        https://stackoverflow.com/questions/43892506/opencv-python-rotate-image-without-cropping-sides
        """
        a = 1.0   # TODO: idk why is 1.0
        height, width = self._shape
        image_center = (width/2, height/2)
        # getRotationMatrix2D needs coordinates in reverse
        # order (width, height) compared to shape
        rot_mat= cv2.getRotationMatrix2D(
            image_center,
            self._rotation_angle,
            a)
        # rotation calculates the cos and sin, taking absolutes of those.
        abs_cos = abs(rot_mat[0, 0])
        abs_sin = abs(rot_mat[0, 1])
        # find the new width and height bounds
        bound_w = int(height * abs_sin + width * abs_cos)
        bound_h = int(height * abs_cos + width * abs_sin)
        # subtract old image center (bringing image back to origo) and adding
        # the new image center coordinates
        rot_mat[0, 2] += bound_w/2 - image_center[0]
        rot_mat[1, 2] += bound_h/2 - image_center[1]
        self._bound = (bound_w, bound_h)
        return rot_mat

    @staticmethod
    def _get_sample_image(config_path: str, vid_identifier: str) -> np.ndarray:
        loader = get_loader(config_path, vid_identifier)
        image = loader.read()
        loader.end()
        return image

    @staticmethod
    def _gray(image: np.ndarray) -> np.ndarray:
        if len(image.shape) == 2:
            return image
        return cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

    def show_entire_image(self):
        """Set slice to cover the entire image"""
        w_slice = slice(0, 6000)
        h_slice = slice(0, 6000)
        self._slice = (w_slice, h_slice)

    def _rotate(self, image: np.ndarray) -> np.ndarray:
        if self._rotation_angle != 0:
            # rotate image with the new bounds and translated rotation matrix
            rotated_mat = cv2.warpAffine(
                image,
                self._rotation_matrix,
                self._bound
                )
            return rotated_mat
        return image
    def _pre_crop(self, image: np.ndarray) -> np.ndarray:
        new_image =  image[self._pre_slice[0], self._pre_slice[1]]
        self._shape = (new_image.shape[0], new_image.shape[1])
        # TODO: this shouldn't be done here. Find a better way
        self._rotation_matrix = self._get_rotation_matrix()
        return new_image

    def _crop(self, image: np.ndarray) -> np.ndarray:
        new_image =  image[self._slice[0], self._slice[1]]
        self._shape = (new_image.shape[0], new_image.shape[1])
        # TODO: this shouldn't be done here. Find a better way
        self._rotation_matrix = self._get_rotation_matrix()
        return new_image

    def apply_roi_extraction(self, image: np.ndarray, gray=True) -> np.ndarray:
        """Apply image rotation, cropping and rgb2gray"""
        # it must be in this order in order to calibrate easier
        image = self._pre_crop(image)
        image = self._rotate(image)
        image = self._crop(image)
        if gray:
            image = self._gray(image)
        return image

    def apply_image_enhancement(self, image: np.ndarray) -> np.ndarray:
        """Apply contrast- and gamma correction"""
        # img_grey = ip.color_corr(
        #     img_orth,
        #     alpha=self.enhance_alpha,
        #     beta=self.enhance_beta,
        #     gamma=self.enhance_gamma)
        return image

    def _crop_using_refs(self, image: np.ndarray) -> np.ndarray:
        image = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
        xs = self._config["dataset"]['gcp']['pixels']['y']
        ys = self._config["dataset"]['gcp']['pixels']['x']
        xslice = slice(min(xs), max(xs)+1)
        yslice = slice(min(ys), max(ys)+1)
        image = image[xslice, yslice]
        self._shape = (image.shape[0], image.shape[1])
        self._or_params = self._get_orthorectification_params(image, reduce=(min(ys), min(xs)))
        self._rotation_matrix = self._get_rotation_matrix()
        return image

    def apply_distortion_correction(self, image: np.ndarray) ->np.ndarray:
        """Given GCP, undistort image."""
        if not self._config["dataset"]['gcp']['apply']:
            return image

        image = self._crop_using_refs(image)
        # apply lens distortion correction
        if self._config["preprocessing"]['image_correction']['apply']:
            image = ip.lens_corr(
                    image,
                    k1=self._config["preprocessing"]['image_correction']['k1'],
                    c=self._config["preprocessing"]['image_correction']['c'],
                    f=self._config["preprocessing"]['image_correction']['f']
                    )

        # apply orthorectification
        image = ip.orthorect_trans(
            image,
           self._or_params[0],
           self._or_params[1]
        )
        self._shape = (image.shape[0], image.shape[1])
        # update rotation matrix such as the shape of the image changed
        self._rotation_matrix = self._get_rotation_matrix()
        return image


def main(config_path: str, video_identifier: str, save_image: bool):
    """Demonstrate basic example of video correction."""
    t0 = time.process_time()
    loader = get_loader(config_path, video_identifier)
    t1 = time.process_time()
    formatter = Formatter(config_path, video_identifier)
    t2 = time.process_time()
    image = loader.read()
    t3 = time.process_time()
    image = formatter.apply_distortion_correction(image)
    t4 = time.process_time()
    image = formatter.apply_roi_extraction(image)
    t5 = time.process_time()
    loader.end()
    t6 = time.process_time()
    print('- get_loader:', t1 - t0)
    print('- Formatter:', t2 - t1)
    print('- loader.read:', t3 - t2)
    print('- formatter.apply_distortion_correction:', t4 - t3)
    print('- formatter.apply_roi_extraction:', t5 - t4)
    print('- loader.end:', t6 - t5)

    if save_image:
        cv2.imwrite('tmp.jpg', image)
    else:
        cv2.imshow('image', cv2.resize(image, (1000, 1000)))
        cv2.waitKey(0)
        cv2.destroyAllWindows()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "statio_name",
        help="Name of the station to be analyzed")
    parser.add_argument(
        "video_identifier",
        help="Index of the video of the json config file")
    parser.add_argument(
        '-s',
        '--save',
        help='Save images instead of showing',
        action='store_true')
    parser.add_argument(
        '-p',
        '--path',
        help='Path to the config folder',
        type=str,
        default=FOLDER_PATH)

    args = parser.parse_args()
    CONFIG_PATH = f'{args.path}/{args.statio_name}.json'
    main(
        config_path=CONFIG_PATH,
        video_identifier=args.video_identifier,
        save_image=args.save
    )



"""Water Level Detector using method described in.
Embedded implementation of image-based water-level measurement system
by:
- Kim, J.
- Han, Y.
- Hahn, H.
"""

import os
import json
import argparse
import numpy as np
import cv2
from loader import get_loader


FOLDER_PATH = '/home/joseph/Documents/Thesis/Dataset/config'


class WaterlevelDetector:
    """Detect water level."""

    def __init__(self, config_path: str, video_identifier: str):
        """Initialize."""
        with open(config_path) as json_file:
            config = json.load(json_file)[video_identifier]['water_level']
        self._loader = get_loader(config_path, video_identifier)
        self._buffer_length = config['buffer_length']
        roi  = config['roi']
        roi2  = config['roi2']
        self._wr0 = slice(roi2[0][0], roi2[1][0])
        self._wr1 = slice(roi2[0][1], roi2[1][1])
        self._r0 = slice(roi[0][0], roi[1][0])
        self._r1 = slice(roi[0][1], roi[1][1])
        self._roi_shape = (roi[1][0] - roi[0][0], roi[1][1] - roi[0][1])

        ksize = config['kernel_size']
        self._kernel = np.ones((ksize, ksize),np.uint8)

    def _get_difference_accumulation(self, plot):
        cnt = 0
        buffer = []
        accumulated_image = np.zeros(self._roi_shape)

        image = self._loader.read()
        ref_image = image[self._wr0, self._wr1]
        np.save('im_ref.npy', ref_image)
        image = image[self._r0, self._r1]
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        np.save('acc0.npy', image)
        image = cv2.medianBlur(image, 5)

        for i in range(self._buffer_length):
            if not self._loader.has_images():
                print('broke')
                return None
            new_image = self._loader.read()[self._r0, self._r1]
            new_image = cv2.cvtColor(new_image, cv2.COLOR_BGR2GRAY)
            new_image = cv2.medianBlur(new_image, 5)

            accumulated_image += (new_image - image) ** 2
            image = new_image

        np.save('acc.npy', accumulated_image)
        if plot:
            os.system("plotNpy acc.npy acc0.npy im_ref.npy")
        d = accumulated_image.mean()
        print('idpp:', round(d, 2))

        return d

    @staticmethod
    def _get_threshold(image):
        """Get threshold."""
        hist, _ = np.histogram(image.ravel(), density=True, bins=255)
        max_idx = 0
        max_slope = abs(hist[0] - hist[1])
        for i in range(len(hist)-1):
            new_slope = abs(hist[i] - hist[i-1])
            if new_slope > max_slope:
                max_slope = new_slope
                max_idx = i
        threshold = int(255 * (max_idx + 1) / len(hist))
        return threshold, hist

    def _compute_water_level(self, image, threshold):
        """Use given threshold compute the water level of the image."""
        print('threshold:', threshold)
        image = (image > threshold).astype(np.uint8)
        np.save('out1.npy', image)

        image = cv2.morphologyEx(image, cv2.MORPH_CLOSE, self._kernel)
        total = image.shape[0] * image.shape[1]
        print('total', total)
        msum = image.sum()
        print('msum', msum)
        p = round(msum / total, 2)
        print('percentage:', p)
        height = image.shape[0] - int(p * image.shape[0])
        for i in range(image.shape[1]):
            image[height][i] = 3
        np.save('out2.npy', image)
        return height

    def get_water_level(self, plot):
        """Calculate and return water level."""
        return self._get_difference_accumulation(plot)
        # np.save('out0.npy', accumulated_image)
        # threshold, _ = self._get_threshold(accumulated_image)
        # height = self._compute_water_level(accumulated_image, threshold)

        # return height


def main(
    config_path: str,
    video_identifier: str,
    show_image=False,
    plot=False
) -> None:
    """Execute basic example of water level detector."""
    water_level_detector = WaterlevelDetector(config_path, video_identifier)
    idpp = water_level_detector.get_water_level(plot)
    return idpp
    # print('water level:', water_level)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "statio_name",
        help="Name of the station to be analyzed")
    parser.add_argument(
        "video_identifier",
        help="Index of the video of the json config file")
    parser.add_argument(
        '-p',
        '--path',
        help='Path to the config folder',
        type=str,
        default=FOLDER_PATH)
    parser.add_argument(
        '-P',
        '--plot',
        action='store_true',
        help='Plot output image')
    args = parser.parse_args()
    CONFIG_PATH = f'{args.path}/{args.statio_name}.json'
    main(
        config_path=CONFIG_PATH,
        video_identifier=args.video_identifier,
        plot=args.plot,
    )



"""Loader of videos of frames."""

import abc
import argparse
import json
import os
from pathlib import Path
from typing import Iterable

import cv2
import numpy as np
from numpy.typing import NDArray

from awive.config import ConfigDataset

FOLDER_PATH = '/home/joseph/Documents/Thesis/Dataset/config'


class Loader(metaclass=abc.ABCMeta):
    """Abstract class of loader."""

    def __init__(self, config: ConfigDataset) -> None:
        """Initialize loader."""
        self._offset: int = config.image_number_offset
        self._index: int = 0
        self.config = config
        self.fps: int = 1
        self.total_frames = 0

    @property
    def image_shape(self):
        """Return the shape of the images."""
        return (self.config.width, self.config.height)

    @property
    def index(self) -> int:
        """Index getter."""
        return self._index

    @abc.abstractmethod
    def has_images(self) -> bool:
        """Check if the source contains one more frame."""

    @abc.abstractmethod
    def read(self) -> NDArray[np.uint8]:
        """Read a new image from the source."""

    @abc.abstractmethod
    def end(self) -> None:
        """Free all resources."""


class ImageLoader(Loader):
    """Loader that loads images from a directory."""

    def __init__(self, config: ConfigDataset) -> None:
        """Initialize loader."""
        super().__init__(config)
        self._image_dataset = config.image_dataset
        self._prefix = config.image_path_prefix
        self._digits = config.image_path_digits
        self._image_number = len(os.listdir(self._image_dataset))

    def has_images(self) -> bool:
        """Check if the source contains one more frame."""
        return self._index < self._image_number

    def _path(self, i: int) -> str:
        i += self._offset
        if self._digits == 5:
            return f'{self._image_dataset}/{self._prefix}{i:05}.jpg'
        if self._digits == 3:
            return f'{self._image_dataset}/{self._prefix}{i:03}.jpg'
        return f'{self._image_dataset}/{self._prefix}{i:04}.jpg'

    def set_index(self, index: int) -> None:
        """Set index of the loader to read any image from the folder."""
        self._index = index

    def read(self) -> np.ndarray:
        """Read a new image from the source."""
        self._index += 1
        path: str = self._path(self._index)
        if not Path(path).exists():
            raise FileNotFoundError(f'Image not found: {path}')
        return cv2.imread(self._path(self._index))

    def read_iter(self) -> Iterable[np.ndarray]:
        """Read a new image from the source."""
        self._index += 1
        path: str = self._path(self._index)
        if Path(path).exists():
            yield cv2.imread(self._path(self._index))

    def end(self) -> None:
        """Free all resources."""
        pass


class VideoLoader(Loader):
    """Loader that loads from a video."""

    def __init__(self, config: ConfigDataset) -> None:
        """Initialize loader."""
        super().__init__(config)

        # check if config.video_path exists
        if not Path(config.video_path).exists():
            raise FileNotFoundError(f'Video not found: {config.video_path}')

        self._cap: cv2.VideoCapture = cv2.VideoCapture(self.config.video_path)
        self._image = None  # Current image
        self._image_read: bool = False  # Check if the current images was read

        # Get number of frames
        cap: cv2.VideoCapture = cv2.VideoCapture(self.config.video_path)
        property_id: int = int(cv2.CAP_PROP_FRAME_COUNT)
        self.total_frames = int(cv2.VideoCapture.get(cap, property_id)) + 1
        self.fps = cap.get(cv2.CAP_PROP_FPS)
        self.width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        self.height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))

        # Skip offset
        for _ in range(self._offset+1):
            if self.has_images():
                self.read()

    def has_images(self):
        """Check if the source contains one more frame."""
        if not self._cap.isOpened():
            return False
        ret, self._image = self._cap.read()
        self._image_read = False
        return ret

    def read(self) -> NDArray[np.uint8]:
        """Read a new image from the source."""
        self._index += 1
        if self._image_read:
            ret, self._image = self._cap.read()
            if not ret:
                print('error at reading')
        self._image_read = True
        return self._image

    def end(self) -> None:
        """Free all resources."""
        self._cap.release()


def make_loader(config: ConfigDataset):
    """Make a loader based on config."""
    image_folder_path = config.image_dataset
    # check if the image_folder_path contains any jpg or png file
    for file in Path(image_folder_path).iterdir():
        if file.suffix in ('.jpg', '.png'):
            return ImageLoader(config)

    return VideoLoader(config)


def get_loader(config_path: str, video_identifier: str) -> Loader:
    """Return a ImageLoader or VideoLoader class.

    :return image_loader: if the image_dataset has any image jpg or png
    :return video_loader: if the previous assumption is not true
    """
    # check if in image folder there are located the extracted images
    config = ConfigDataset(
        **json.loads(
            Path(config_path).read_text()
        )[video_identifier]['dataset']
    )
    return make_loader(config)


def main(config_path: str, video_identifier: str, save_image: bool):
    """Execute a basic example of loader."""
    loader = get_loader(config_path, video_identifier)
    image = loader.read()
    if save_image:
        cv2.imwrite('tmp.jpg', image)
    else:
        cv2.imshow('image', cv2.resize(image, (1000, 1000)))
        cv2.waitKey(0)
        cv2.destroyAllWindows()

    loader.end()


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "statio_name",
        help="Name of the station to be analyzed")
    parser.add_argument(
        "video_identifier",
        help="Index of the video of the json config file")
    parser.add_argument(
        '-s',
        '--save',
        help='Save images instead of showing',
        action='store_true')
    parser.add_argument(
        '-p',
        '--path',
        help='Path to the config folder',
        type=str,
        default=FOLDER_PATH)

    args = parser.parse_args()
    CONFIG_PATH = f'{args.path}/{args.statio_name}.json'
    main(
        config_path=CONFIG_PATH,
        video_identifier=args.video_identifier,
        save_image=args.save
    )



"""Main application of Adaptive Water Image Velocimetry Estimator."""

import argparse

import cv2
import numpy as np
from numpy.typing import NDArray

from awive.loader import Loader, get_loader


DEFAULT_DATASET_PATH = "/home/joseph/Documents/Thesis/Dataset/config"


def main(config_path: str, video_identifier: str) -> None:
    """Use basic loader functions to read an image."""
    loader: Loader = get_loader(config_path, video_identifier)
    image: NDArray[np.uint8] = loader.read()
    cv2.imshow('image', image)
    cv2.waitKey(0)
    cv2.destroyAllWindows()
    loader.end()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "statio_name",
        help="Name of the station to be analyzed"
    )
    parser.add_argument(
        "video_identifier",
        help="Index of the video of the json config file"
    )
    parser.add_argument(
        "--dataset_path",
        dest="dataset_path",
        default=DEFAULT_DATASET_PATH,
        help="Path to the dataset folder",
    )
    args = parser.parse_args()
    main(
        config_path=f"{args.dataset_path}/{args.statio_name}.json",
        video_identifier=args.video_identifier,
    )



"""Optical Tracking Image Velocimetry."""

import argparse
import random
import json
import math
import numpy as np
import cv2
from awive.correct_image import Formatter
from awive.loader import (get_loader, Loader)
# from sti import applyMean


FOLDER_PATH = "examples/datasets"


def get_magnitude(kp1, kp2):
    """Get the distance between two keypoints."""
    # return math.sqrt((kp2.pt[1] - kp1.pt[1])**2 + (kp2.pt[0] - kp1.pt[0]) ** 2 )
    return abs(kp2.pt[0] - kp1.pt[0])


def get_angle(kp1, kp2):
    """Get angle between two key points."""
    return math.atan2(kp2.pt[1] - kp1.pt[1] , kp2.pt[0] - kp1.pt[0]) *  180 / math.pi


def _get_velocity(kp1, kp2, real_distance_pixel, time, fps):
    if time == 0:
        return 0
    return get_magnitude(kp1, kp2) * real_distance_pixel * fps / time


def reject_outliers(data, m=2.):
    d = np.abs(data - np.median(data))
    mdev = np.median(d)
    s = d/mdev if mdev else 0.
    return data[s<m]


def compute_stats(velocity, hist=False):
    v = np.array(sum(velocity, []))
    v = reject_outliers(v)
    count = len(v)
    if count == 0:
        return 0, 0, 0, 0, 0
    avg = v.mean()
    max_ = v.max()
    min_ = v.min()
    std_dev = np.std(v)

    if hist:
        import matplotlib.pyplot as plt
        plt.hist(v.astype(int))
        plt.ylabel('Probability')
        plt.xlabel('Data');
        plt.show()

    return avg, max_, min_, std_dev, count


class OTV():
    """Optical Tracking Image Velocimetry."""

    def __init__(
        self,
        config_path: str,
        video_identifier: str,
        prev_gray: np.ndarray,
        debug=0
    ) -> None:
        self._debug = debug
        with open(config_path) as json_file:
            root_config = json.load(json_file)[video_identifier]
            config = root_config['otv']
        self._partial_max_angle = config['partial_max_angle']
        self._partial_min_angle = config['partial_min_angle']
        self._final_max_angle = config['final_max_angle']
        self._final_min_angle = config['final_min_angle']
        self._final_min_distance = config['final_min_distance']
        self._max_features = config['max_features']
        self._radius = config['lk']['radius']
        self._max_level = config['lk']['max_level']
        self._step = config['region_step']
        self._resolution = config['resolution']
        self._pixel_to_real = config['pixel_to_real']

        self._width = root_config["preprocessing"]['roi']['w2']  -root_config["preprocessing"]['roi']['w1']
        self._height = root_config["preprocessing"]['roi']['h2'] - root_config["preprocessing"]['roi']['h1']
        mask_path = config['mask_path']
        self._regions = config['lines']
        if len(mask_path) != 0:
            self._mask = cv2.imread(mask_path, 0) > 1
            self._mask = cv2.resize(
                self._mask.astype(np.uint8),
                (self._height, self._width),
                cv2.INTER_NEAREST
            )
        else:
            self._mask = None

        winsize = config['lk']['winsize']

        self.lk_params = {
            'winSize': (winsize, winsize),
            'maxLevel': self._max_level,
            'criteria': (cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT,
                         config['lk']['max_count'],
                         config['lk']['epsilon']),
            'flags': cv2.OPTFLOW_LK_GET_MIN_EIGENVALS,
            'minEigThreshold': config['lk']['min_eigen_threshold']
            }
        self.prev_gray = prev_gray

    def _partial_filtering(self, kp1, kp2, max_distance):
        magnitude = get_magnitude(kp1, kp2)  #only to limit the research window
        if magnitude > max_distance:
            return False
        angle = get_angle(kp1, kp2)
        if angle < 0:
            angle = angle + 360
        if angle == 0 or angle == 360:
            return True
        if self._partial_min_angle <= angle <= self._partial_max_angle:
            return True
        return False

    def _final_filtering(self, kp1, kp2):
        """Final filter of keypoints"""
        magnitude = get_magnitude(kp1, kp2)
        if magnitude < self._final_min_distance:
            return False
        angle = get_angle(kp1, kp2)
        if angle < 0:
            angle = angle + 360
        if angle == 0 or angle == 360:
            return True
        if self._final_min_angle <= angle <= self._final_max_angle:
            return True
        return False


    def _apply_mask(self, image):
        if self._mask is not None:
            image = image * self._mask
        return image

    def _init_subregion_list(self, dimension, width):
        ret = []
        n_regions = math.ceil(width / self._step)
        for _ in range(n_regions):
            # TODO: This is so inneficient
            if dimension == 1:
                ret.append(0)
            elif dimension == 2:
                ret.append([])
        return ret


    def run(self, loader: Loader, formatter: Formatter, show_video=False):
        """Execute OTV and get velocimetry"""
        # initialze parametrers
        detector = cv2.FastFeatureDetector_create()
        previous_frame = None
        keypoints_current = []
        keypoints_start = []
        time = []
        keypoints_predicted = []
        masks = []

        valid = []
        velocity_mem = []
        keypoints_mem_current = []
        keypoints_mem_predicted = []
        velocity = []
        angle = []
        distance = []
        path = []
        traj_map = np.zeros((1200, 900))

        # update width and height if needed
        if loader.image_shape[0] < self._width:
            self._width = loader.image_shape[0]
        if loader.image_shape[1] < self._height:
            self._height = loader.image_shape[1]


        subregion_velocity = self._init_subregion_list(2, self._width)
        subregion_trajectories = self._init_subregion_list(1, self._width)

        regions = list([] for _ in range(len(self._regions)))

        # Initialization
        for i in range(loader.total_frames):
            valid.append([])
            velocity_mem.append([])
            velocity.append([])
            angle.append([])
            distance.append([])
            path.append([])

        while loader.has_images():
            # get current frame
            current_frame = loader.read()
            current_frame = formatter.apply_distortion_correction(current_frame)
            current_frame = formatter.apply_roi_extraction(current_frame)
            # current_frame = self._apply_mask(current_frame)

            # get features as a list of KeyPoints
            keypoints = list(detector.detect(current_frame, None))
            # print(f"{type(keypoints)=}")
            # print(f"{len(keypoints)=}")
            # print(f"{type(keypoints[0])=}")
            random.shuffle(keypoints)

            # update lot of lists
            if previous_frame is None:
                for i, keypoint in enumerate(keypoints):
                    if len(keypoints_current) < self._max_features:
                        keypoints_current.append(keypoint)
                        keypoints_start.append(keypoint)
                        time.append(loader.index)
                        valid[loader.index].append(False)
                        velocity_mem[loader.index].append(0)
                        path[loader.index].append(i)
            else:
                for i, keypoint in enumerate(reversed(keypoints)):
                    if len(keypoints_current) < self._max_features:
                        keypoints_current.append(keypoint)
                        keypoints_start.append(keypoint)
                        time.append(loader.index)
                        valid[loader.index].append(False)
                        velocity_mem[loader.index].append(0)

            if self._debug >= 1:
                print('Analyzing frame:', loader.index)
            if previous_frame is not None:
                pts1 = cv2.KeyPoint_convert(keypoints_current)
                pts2, st, err = cv2.calcOpticalFlowPyrLK(
                    previous_frame,
                    current_frame,
                    pts1,
                    None,
                    **self.lk_params
                    )

                # add predicted by Lucas-Kanade new keypoints
                keypoints_predicted.clear()
                for pt2 in pts2:
                    keypoints_predicted.append(cv2.KeyPoint(
                        pt2[0],
                        pt2[1],
                        1.0
                        ))

                max_distance = self._max_level * (2 * self._radius + 1)
                max_distance /= self._resolution

                k = 0

                for i, keypoint in enumerate(keypoints_current):
                    partial_filter = self._partial_filtering(
                        keypoint,
                        keypoints_predicted[i],
                        max_distance
                        )
                    # check if the trajectory finished or the vector is invalid
                    if not (st[i] and partial_filter):
                        final_filter = self._final_filtering(
                            keypoints_start[i],
                            keypoints_current[i]
                            )
                        # check if it is a valid trajectory
                        if final_filter:
                            velocity_i = _get_velocity(
                                keypoints_start[i],
                                keypoints_current[i],
                                self._pixel_to_real / self._resolution,
                                loader.index - time[i],
                                loader.fps
                                )
                            angle_i = get_angle(
                                keypoints_start[i],
                                keypoints_current[i]
                                )

                            xx0 = int(keypoints_start[i].pt[1])
                            yy0 = int(keypoints_start[i].pt[0])
                            traj_map[xx0][yy0] += 100
                            # sub-region computation
                            # module_start = int(keypoints_start[i].pt[1] /
                            #         self._step)
                            # module_current = int(keypoints_current[i].pt[1] /
                            #         self._step)
                            # if module_start == module_current:
                            # subregion_velocity[module_start].append(velocity_i)
                            # subregion_trajectories[module_start] += 1

                            for r_idx, region in enumerate(self._regions):
                                if abs(xx0 - region) < 15:
                                    regions[r_idx].append(velocity_i)

                            # update storage
                            pos = i
                            j = loader.index - 1
                            while j >= time[i]:
                                valid[j][pos] = True
                                velocity_mem[j][pos] = velocity_i
                                pos = path[j][pos]
                                j-=1

                            velocity[loader.index].append(velocity_i)
                            angle[loader.index].append(angle_i)
                            distance[loader.index].append(
                                velocity_i * (loader.index - time[i]) / loader.fps)

                        continue

                    # Add new displacement vector
                    keypoints_current[k] = keypoints_current[i]
                    keypoints_start[k] = keypoints_start[i]
                    keypoints_predicted[k] = keypoints_predicted[i]
                    path[loader.index].append(i)
                    velocity_mem[loader.index].append(0)
                    valid[loader.index].append(False)
                    time[k] = time[i]
                    k += 1

                # Only keep until the kth keypoint in order to filter invalid
                # vectors
                keypoints_current = keypoints_current[:k]
                keypoints_start = keypoints_start[:k]
                keypoints_predicted = keypoints_predicted[:k]
                time = time[:k]


            if self._debug >= 1:
                print('number of trajectories:', len(keypoints_current))

            if show_video:
                if previous_frame is not None:
                    color_frame = cv2.cvtColor(current_frame, cv2.COLOR_GRAY2RGB)
                    output = draw_vectors(
                        color_frame,
                        keypoints_predicted,
                        keypoints_current,
                        masks
                        )
                    cv2.imshow("sparse optical flow", output)
                    if cv2.waitKey(10) & 0xFF == ord('q'):
                        break
            previous_frame = current_frame.copy()
            keypoints_mem_current.append(keypoints_current)
            keypoints_mem_predicted.append(keypoints_predicted)

            # TODO: I guess the swap is not needed such as in the next iteration
            # the keypoints_predicted will be cleaned
            if len(keypoints_predicted) != 0:
                keypoints_predicted, keypoints_current = keypoints_current, keypoints_predicted
        np.save('traj.npy', traj_map)

        loader.end()
        if show_video:
            cv2.destroyAllWindows()
        avg, max_, min_, std_dev, count = compute_stats(velocity, show_video)

        if self._debug >= 1:
            print('avg:', round(avg, 4))
            print('max:', round(max_, 4))
            print('min:', round(min_, 4))
            print('std_dev:', round(std_dev, 2))
            print('count:', count)

        out_json = {}
        for i, sv in enumerate(regions):
            out_json[str(i)] = {}
            t = np.array(sv)
            t = t[t!=0]
            if len(t) != 0:
                t = reject_outliers(t)
                m = t.mean()
            else:
                m = 0
            out_json[str(i)]['velocity'] = m
            out_json[str(i)]['count'] = len(t)
        return out_json


def draw_vectors(image, new_list, old_list, masks):
    """Draw vectors of velocity and return the output and update mask"""
    if len(image.shape) == 3:
        color = (0, 255, 0)
        thick = 1
    else:
        color = 255
        thick = 1

    # create new mask
    mask = np.zeros(image.shape, dtype=np.uint8)
    for new, old in zip(new_list, old_list):
        # TODO: Perhaps there is a more efficient way to transform this to a
        # numpy array
        new_pt = np.array([int(new.pt[0]), int(new.pt[1])])
        old_pt = np.array([int(old.pt[0]), int(old.pt[1])])
        mask = cv2.line(mask, new_pt.ravel(), old_pt.ravel(), color, thick)

    # update masks list
    masks.append(mask)
    if len(masks) < 3:
        return np.zeros(image.shape)
    if len(masks) > 3:
        masks.pop(0)


    # generate image with mask
    total_mask = np.zeros(mask.shape, dtype=np.uint8)
    for mask_ in masks:
        total_mask = cv2.add(total_mask, mask_)
    output = cv2.add(image, total_mask)
    return output


def main(config_path: str, video_identifier: str, show_video=False, debug=0):
    """Basic example of OTV"""
    loader = get_loader(config_path, video_identifier)
    formatter = Formatter(config_path, video_identifier)
    loader.has_images()
    image = loader.read()
    prev_gray = formatter.apply_distortion_correction(image)
    prev_gray = formatter.apply_roi_extraction(prev_gray)
    otv = OTV(config_path, video_identifier, prev_gray, debug)
    return otv.run(loader, formatter, show_video)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "statio_name",
        help="Name of the station to be analyzed"
    )
    parser.add_argument(
        "video_identifier",
        help="Index of the video of the json config file"
    )
    parser.add_argument(
        '-p',
        '--path',
        help='Path to the config folder',
        type=str,
        default=FOLDER_PATH
    )
    parser.add_argument(
        '-d',
        '--debug',
        help='Activate debug mode',
        type=int,
        default=0
    )
    parser.add_argument(
        '-v',
        '--video',
        action='store_true',
        help='Play video while processing'
    )
    args = parser.parse_args()
    print(
        main(
            config_path=f'{args.path}/{args.statio_name}/config.json',
            video_identifier=args.video_identifier,
            show_video=args.video,
            debug=args.debug
        )
    )



"""Class ImageVelocimetry."""

import abc
from awive.loader import Loader


class ImageVelocimetry(abc.ABC):
    """Abstract class for image velocimetry algorithms."""

    def __init__(self) -> None:
        """Initialize the class."""

    @abc.abstractmethod
    def run(self, loader: Loader) -> None:
        """Run the algorithm."""
        raise NotImplementedError



"""Draw a circle in the ground control point."""

import argparse
import json

import cv2
import numpy as np

from awive.correct_image import Formatter
from libs.npyplotter.npyplotter.plot_npy import picshow


FOLDER_PATH = '/home/joseph/Documents/Thesis/Dataset/config'
TEXT_FACE = cv2.FONT_HERSHEY_DUPLEX
TEXT_SCALE = 0.5
TEXT_THICKNESS = 2


def draw_circle(img, center, number):
    """Draw a circle with a number."""
    text = str(number)
    cv2.circle(img, center, 10, 255, -1)

    text_size, _ = cv2.getTextSize(text, TEXT_FACE, TEXT_SCALE, TEXT_THICKNESS)
    text_origin: tuple[int, int] = (
        int(center[0] - text_size[0] / 2),
        int(center[1] + text_size[1] / 2)
    )

    cv2.putText(
        img,
        text,
        text_origin,
        TEXT_FACE,
        TEXT_SCALE,
        0,
        TEXT_THICKNESS,
        cv2.LINE_AA
    )


def main(
    config_path: str,
    video_identifier: str,
    plot=False,
    undistort=False,
    roi=False
) -> None:
    """Read config and draw circles in the image."""
    with open(config_path) as json_file:
        config = json.load(json_file)[video_identifier]['gcp']
    formatter = Formatter(config_path, video_identifier)

    img = np.load('tmp.npy')

    pixels = config['pixels']

    le = len(pixels['x'])
    for i in range(le):
        draw_circle(img, (pixels['x'][i], pixels['y'][i]), i+1)

    if undistort:
        img = formatter.apply_distortion_correction(img)
    if roi:
        img = formatter.apply_roi_extraction(img)

    np.save('tmp2.npy', img)
    if plot:
        picshow([img])


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "statio_name",
        help="Name of the station to be analyzed"
    )
    parser.add_argument(
        "video_identifier",
        help="Index of the video of the json config file"
    )
    parser.add_argument(
        '-p',
        '--path',
        help='Path to the config folder',
        type=str,
        default=FOLDER_PATH
    )
    parser.add_argument(
        '-P',
        '--plot',
        action='store_true',
        help='Plot output image'
    )
    parser.add_argument(
        '-r',
        '--roi',
        action='store_true',
        help='Format image using selecting only roi area'
    )
    parser.add_argument(
        '-u',
        '--undistort',
        action='store_true',
        help='Format image using distortion correction'
    )
    args = parser.parse_args()
    main(
        config_path=f'{args.path}/{args.statio_name}.json',
        video_identifier=args.video_identifier,
        undistort=args.undistort,
        roi=args.roi,
        plot=args.plot
    )



"""Read csv."""
import pandas as pd
import matplotlib.cm
import cv2 as cv
from awive.utils import ImageLoader

path = ('/home/joseph/Documents/Thesis/Dataset/TowardHarmonization/dataset/'
        'CastorRiver/20190410/Reference/ref5.csv')

image_dataset = ('/home/joseph/Documents/Thesis/Dataset/TowardHarmonization/'
                 'dataset/CastorRiver/20190410/Video3/Frames')
image_len =  565
image_offset = 51
resize_ratio = 5

def get_df():
    df = pd.read_csv(path)
    df = df.drop(['d'], axis=1)
    df = df[df['v'].notna()]
    df['v'] = df['v'].str.replace(',', '.').astype(float)
    df['v'] = pd.to_numeric(df['v'])
    return df


def main():
    df = get_df()
    image_loader = ImageLoader(image_dataset, '', 5, image_offset)
    im = image_loader.load(0)
    im = cv.cvtColor(im, cv.COLOR_GRAY2RGB)
    cmap = matplotlib.cm.get_cmap('Spectral')
    minVal = df['v'].min()
    maxVal = df['v'].max()
    rangeVal = maxVal - minVal

    for i, val in df.iterrows():
        v =(val['v']-minVal)/rangeVal
        start_point = (int(val['x']), int(val['y']))
        maxI = 400
        end_point = (int(val['x'] - maxI * v), int(val['y'] - maxI * v))
        rgb = cmap(v)
        color = (255*rgb[0], 255*rgb[1], 255*rgb[2])
        thickness = 20
        im = cv.arrowedLine(im, start_point, end_point, color, thickness)


    width = im.shape[0]
    height = im.shape[1]
    # M = cv.getRotationMatrix2D((width//2, height//2), 44, 1.0)
    M = cv.getRotationMatrix2D((width//2, height//2), 44+180, 1.0)
    im = cv.warpAffine(im, M, (width, height))
    im = im[:2880, 2268:4447]

    new_shape = (width // resize_ratio, height // resize_ratio)
    im = cv.resize(im, new_shape)
    cv.imshow('main', im)
    cv.waitKey(0)
    cv.destroyAllWindows()


if __name__ == "__main__":
    main()



"""Execute both methods."""
import argparse
import time

import otv
import sti


FOLDER_PATH = '/home/joseph/Documents/Thesis/Dataset/config'


def execute_method(method_name, station_name, video_identifier, config_path):
    """Execute a method."""
    t = time.process_time()
    if method_name == 'sti':
        ret = sti.main(config_path, video_identifier)
    elif method_name == 'otv':
        ret = otv.main(config_path, video_identifier)
    else:
        ret = None
    elapsed_time = (time.process_time() - t)
    print(method_name)
    for i in range(len(ret)):
        print(ret[str(i)]['velocity'])
    print(elapsed_time)


def main(station_name, video_identifier, config_path):
    """Execute both methods."""
    execute_method('sti', station_name, video_identifier, config_path)
    execute_method('otv', station_name, video_identifier, config_path)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "station_name",
        help="Name of the station to be analyzed")
    parser.add_argument(
        "video_identifier",
        help="Index of the video of the json config file")
    parser.add_argument(
        '-p',
        '--path',
        help='Path to the config folder',
        type=str,
        default=FOLDER_PATH)
    args = parser.parse_args()
    main(
        station_name=args.station_name,
        video_identifier=args.video_identifier,
        config_path = f'{args.path}/{args.station_name}.json',
    )



"""Image preparation."""
import cv2
import numpy as np


def lens_corr(img, k1=-10.0e-6, c=2, f=8.0):
    """Lens distortion correction based on lens charateristics.

    :param img: original image
    :param k1: barrel lens distortion parameter
    :param c: optical center
    :param f: focal length

    :return corr_img: image corrected for lens distortion
    """
    # define imagery charateristics
    height = img.shape[0]
    width = img.shape[1]

    # define distortion coefficient vector
    dist = np.zeros((4, 1), np.float64)
    dist[0, 0] = k1

    # define camera matrix
    mtx = np.eye(3, dtype=np.float32)

    mtx[0, 2] = width/c     # define center x
    mtx[1, 2] = height/c    # define center y
    mtx[0, 0] = f           # define focal length x
    mtx[1, 1] = f           # define focal length y

    # correct image for lens distortion
    corr_img = cv2.undistort(img, mtx, dist)

    return corr_img


def xy_coord(df):
    """
    Turn longitudes and latitudes into XY coordinates using an Equirectangular
    projection. Only applicable on a small scale.

    Input DataFrame as follows:
    df = pd.DataFrame({'lon': [a, b, c, d],
                                       'lat': [e, f, g, h]})

    Input:
    ------
    df - DataFrame containing columns with longitudes and latitudes
    PPM - Pixels per metre

    Output:
    -------
    df_new - DataFrame with xy-coordinates in metres
    """
    # set base parameters
    # r = 6378137  # meters according to WGS84
    # phi_0 = df.latitude[0]
    # cos_phi_0 = math.cos(math.radians(phi_0))

    # # create new DataFrame containing original coordinates in metres
    # df_new = pd.DataFrame()
    # df_new['x'] = [r * math.radians(lon) * cos_phi_0 for lon in df.lon.values]
    # df_new['y'] = [r * math.radians(lat) for lat in df.lat.values]

    # return df_new
    return 12


def orthorect_param(img, df_from, df_to, PPM=100, lonlat=False):
    """Image orthorectification parameters based on 4 GCPs.
    GCPs need to be at water level.

    Input DataFrames as follows:
    df = pd.DataFrame({'x': [a, b, c, d],
                                       'y': [e, f, g, h]})

    Input:
    ------
    img - original image
    df_from - DataFrame containing the xy-coordinates of the GCPs in
            the imagery in pixels
    df_to - DataFrame DataFrame containing the real xy-coordinates of
            the GCPs in metres (if in lon-lat coordinates, set lonlat=True
            to convert)
    PPM - pixels per meter in the corrected imagery (default: 100)
    lonlat - convert longitudes/latitudes to metres (default: False)

    Output:
    -------
    M_new - Transformation matrix based on image corners
    C_new - Coordinates of image corners in the orthorectified imagery
    df_to - Original coordinates of the GCPs in metres (relevant if
            converted from geographic coordinate system)
    """
    if lonlat:
        df_to = xy_coord(df_to)

    # set points to float32
    pts1 = np.float32(df_from)
    # Multiple elements inside df_to by PPM
    df_too = []
    for x in df_to:
        df_too.append(list(map(lambda x: x*PPM, x)))
    pts2 = np.float32(df_too)

    # define transformation matrix based on GCPs
    M = cv2.getPerspectiveTransform(pts1, pts2)

    # find locations of transformed image corners
    # height, width, __ = img.shape
    height = img.shape[0]
    width = img.shape[1]
    C = np.array([[0, 0, 1],
                  [width, 0, 1],
                  [0, height, 1],
                  [width, height, 1]])
    C_new = np.array([(np.dot(i, M.T) / np.dot(i, M.T)[2])[:2] for i in C])

    C_new[:, 0] -= min(C_new[:, 0])
    C_new[:, 1] -= min(C_new[:, 1])

    # define new transformation matrix based on image corners
    # otherwise, part of the imagery will not be saved
    M_new = cv2.getPerspectiveTransform(
        np.float32(C[:, :2]),
        np.float32(C_new)
    )

    return M_new, C_new, df_to


def orthorect_trans(img, M, C):
    """
    Image orthorectification based on parameters found with
    ip.orthorect_param().

    Input:
    ------
    img - original image
    M - Transformation matrix based on image corners
    C - Coordinates of image corners in the orthorectified imagery

    Output:
    -------
    corr_img - orthorectified image
    """

    # define corrected image dimensions based on C
    cols = int(np.ceil(max(C[:, 0])))
    rows = int(np.ceil(max(C[:, 1])))

    # orthorectify image
    corr_img = cv2.warpPerspective(img, M, (cols, rows))

    return corr_img


def color_corr(img, alpha=None, beta=None, gamma=0.5):
    """
    Grey scaling, contrast- and gamma correction. Both alpha and beta need to
    be defined in order to apply contrast correction.

    Input:
    ------
    img - original image
    alpha - gain parameter for contrast correction (default: None)
    beta - bias parameter for contrast correction (default: None)
    gamma - brightness parameter for gamma correction (default: 0.5)

    Output:
    -------
    corr_img - gray scaled, contrast- and gamma corrected image
    """
    # turn image into grey scale
    corr_img = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)

    if alpha and beta:
        # apply contrast correction
        corr_img = cv2.convertScaleAbs(corr_img, alpha=alpha, beta=beta)

    # apply gamma correction
    invGamma = 1./gamma
    table = (np.array([
        ((i / 255.0) ** invGamma) * 255
        for i in np.arange(0, 256)]).astype('uint8'))

    corr_img = cv2.LUT(corr_img, table)

    return corr_img



"""Mean."""
import cv2 as cv
import numpy as np

from matplotlib import pyplot as plt
from utils import ImageLoader

imageDataset = '/home/joseph/Documents/Thesis/Dataset/ssivDataset/images'
imageLength = 372
image_offset = 200


def image_mean():
    imshape = load_image(0).shape
    image_acumulator = np.zeros(imshape, dtype=np.int64)
    for i in range(imageLength):
        image_acumulator += load_image(i)
    return (image_acumulator / imageLength).astype(np.uint8)


def main():
    image_loader = ImageLoader(imageDataset, image_offset)
    imMean = image_mean()
    cv.imshow('image mean', imMean)
    # for i in range(3):
    #     cv.imshow(f'frame: {i:03}', load_image(i) - imMean)
    im = image_loader.load(0)
    for i in range(10, 100, 20):
        s = (imMean*(i/100)).astype(np.uint8)
        cv.imshow(f'k = {i:02}%', im-s)

    cv.imshow('original image', im)
    cv.waitKey(0)
    cv.destroyAllWindows()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['awive', 'awive.algorithms']

package_data = \
{'': ['*']}

install_requires = \
['autopep8>=2.0.0,<3.0.0',
 'bandit>=1.7.4,<2.0.0',
 'flake8>=5.0.4,<6.0.0',
 'isort>=5.10.1,<6.0.0',
 'matplotlib>=3.6.2,<4.0.0',
 'mypy>=0.982,<0.983',
 'opencv-python>=4.6.0.66,<5.0.0.0',
 'pandas>=1.5.1,<2.0.0',
 'pydantic>=1.10.2,<2.0.0',
 'pydocstyle>=6.1.1,<7.0.0',
 'requests>=2.28.1,<3.0.0',
 'types-requests>=2.28.11.5,<3.0.0.0',
 'wget>=3.2,<4.0']

setup_kwargs = {
    'name': 'awive',
    'version': '0.1.0',
    'description': "Joseph Pena's bachelor thesis work",
    'long_description': '# AWIVE\nAWIVE, the acrynomes of Adaptive Water Image Velocimetry Estimator, is a\nsoftware package for the estimation of the velocity field from a sequence of\nimages. It is composed by the methods STIV and OTV, which aim at estimate\nvelocimetry with a low computational cost.\n\n## Installing\n\nInstall and update using pip:\n\n```bash\npip install awive\n```\n\n',
    'author': 'Joseph P.',
    'author_email': 'joseph.pena@utec.edu.pe',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '==3.9.16',
}


setup(**setup_kwargs)



"""Space Time Image Velocimetry."""

import json
import math
import argparse
import cv2
import numpy as np
from correct_image import Formatter
from loader import get_loader
import time


FOLDER_PATH = '/home/joseph/Documents/Thesis/Dataset/config'

cnt = 0


class STIV():
    """Space Time Image Velocimetry."""

    def __init__(self, config_path: str, video_identifier: str, debug=0):
        """Initialize STIV."""
        with open(config_path) as json_file:
            root_config = json.load(json_file)[video_identifier]
            self._config = root_config['stiv']
        self._debug = debug
        # Shall be initialized later
        self._fps = None

        self._stis = []
        self._stis_qnt = len(self._config['lines'])
        self._ksize = self._config['ksize']
        t0 = time.process_time()
        self._generate_st_images(config_path, video_identifier)
        t1 = time.process_time()
        self._overlap = self._config['overlap']

        self._ppm = root_config['PPM']
        if self._debug >= 1:
            print('frames per second:', self._fps)
            print('pixels per meter', self._ppm)

        # create filter window
        w_size = self._config['filter_window']
        W_mn = (1 - np.cos(2 * math.pi * np.arange(w_size) / w_size)) / 2
        W_mn = np.tile(W_mn, (w_size, 1))
        self._filter_win = W_mn * W_mn.T

        # vertical and horizontal filter width
        self._vh_filter = 1

        self._polar_filter_width = self._config['polar_filter_width']
        print('- generate_st_images\t', t1 - t0)

    def _get_velocity(self, angle):
        """Given STI pattern angle, calculate velocity.
        :param angle: angle in radians
        """
        velocity = math.tan(angle) * self._fps / self._ppm
        return velocity

    def _generate_st_images(self, config_path, video_identifier):
        # generate space time images
        loader = get_loader(config_path, video_identifier)
        if self._debug >= 1:
            print('number of frames:', loader.total_frames)
        formatter = Formatter(config_path, video_identifier)
        self._fps = loader.fps

        # initialize set of sti images
        for _ in range(self._stis_qnt):
            self._stis.append([])

        # generate all lines
        coordinates_list = self._config['lines']
        while loader.has_images():
            image = loader.read()
            image = formatter.apply_distortion_correction(image)
            image = formatter.apply_roi_extraction(image)

            for i, coordinates in enumerate(coordinates_list):
                start = coordinates['start']
                end = coordinates['end']
                row = image[start[0], start[1]:end[1]]
                self._stis[i].append(row)

        for i in range(self._stis_qnt):
            self._stis[i] = np.array(self._stis[i])
            np.save(f'images/stiv/sti_{i:04}.npy', self._stis[i])

    @staticmethod
    def _get_main_freqs(isd):
        main_freqs = []
        main_freqs.append(np.argmax(isd))
        if main_freqs[0] < len(isd) / 2:
            main_freqs.append(main_freqs[0] + len(isd) / 2)
        else:
            main_freqs.append(main_freqs[0] - len(isd) / 2)
        return main_freqs

    def _apply_angle(self, mask, freq, isd):
        x = int(freq - self._polar_filter_width)
        y = int(freq + self._polar_filter_width)
        if x < 0:
            x = 0
            mask[x + len(isd):, :] = 1
        elif y > len(isd):
            y = len(isd)
            mask[:y - len(isd), :] = 1
        mask[x:y, :] = 1
        return mask

    def _generate_polar_mask(self, polar_img):
        # calculate Integral Spectrum Distribution
        isd = np.sum(polar_img.T, axis=0)
        main_freqs = self._get_main_freqs(isd)
        mask = np.zeros(polar_img.shape)

        mask = self._apply_angle(mask, main_freqs[0], isd)
        mask = self._apply_angle(mask, main_freqs[1], isd)
        return mask

    @property
    def stis(self):
        """Return Space-Time image."""
        return self._stis

    def _process_sti(self, image: np.ndarray):
        """Process sti image."""
        # image = cv2.medianBlur(image, 5)
        sobelx = cv2.Sobel(image, cv2.CV_64F, 0, 1, ksize=self._ksize)
        sobelt = cv2.Sobel(image, cv2.CV_64F, 1, 0, ksize=self._ksize)
        if sobelx.sum() == 0 and sobelt.sum() == 0:
            if self._debug >= 1:
                print("WARNING: gradients are zero")
            return 0, 0

        Jxx = (sobelx * sobelx).sum()
        Jtt = (sobelt * sobelt).sum()
        Jxt = (sobelx * sobelt).sum()
        angle = math.atan2(2*Jxt, Jtt - Jxx) / 2
        coherence = math.sqrt((Jtt-Jxx)**2 + 4*Jxt**2) / (Jxx + Jtt)
        return angle, coherence

    @staticmethod
    def _get_new_point(point, angle, length):
        """
        point - Tuple (x, y)
        angle - Angle you want your end point at in degrees.
        length - Length of the line you want to plot.

        Will plot the line on a 10 x 10 plot.
        """
        # unpack the first point
        x, y = point
        # find the end point
        endy = length * math.cos(angle)
        endx = length * math.sin(angle)
        return int(endx+x), int(-endy+y)

    def _draw_angle(self, image, angle, position, thick=1, amplitud=10):
        new_point = self._get_new_point(position, angle, amplitud)
        cv2.line(image, new_point, position, 255, thick)
        return image

    @staticmethod
    def _conv2d(a, f):
        """
        https://stackoverflow.com/questions/43086557/convolve2d-just-by-using-numpy
        """
        s = f.shape + tuple(np.subtract(a.shape, f.shape) + 1)
        strd = np.lib.stride_tricks.as_strided
        subM = strd(a, shape = s, strides = a.strides * 2)
        return np.einsum('ij,ijkl->kl', f, subM)

    @staticmethod
    def _to_polar_system(img: np.ndarray, option='convert'):
        """
        Transform 2d image to polar system
        """
        if option == 'invert':
            flag = cv2.WARP_INVERSE_MAP
        else:
            flag = cv2.WARP_FILL_OUTLIERS

        # TODO: I should add padding

        row, col = img.shape
        cent = (int(col / 2), int(row / 2))
        max_radius = int(np.sqrt(row ** 2 + col ** 2) / 2)
        polar = cv2.linearPolar(img, cent, max_radius, flag)
        return polar

    def _filter_sti(self, sti: np.ndarray):
        """
        Filter image using method proposed in:
        "An improvement of the Space-Time Image Velocimetry combined with a new
        denoising method for estimating river discharge"
        by:
        - Zhao, Haoyuan
        - Chen, Hua
        - Liu, Bingyi
        - Liu, Weigao
        - Xu, Chong Yu
        - Guo, Shenglian
        - Wang, Jun
        """
        # resize in order to have more precision
        x = min(sti.shape)
        if x == sti.shape[0]:
            sti = sti[:,:x]
        else:
            sti = sti[:x,:]
        if self._debug >= 1:
            print('size before reshape:', x)
        # the example of the paper uses 600x600, so do I
        sti = cv2.resize(sti, (600, 600), interpolation=cv2.INTER_LINEAR)
        np.save(f'images/stiv/f_{cnt}_0.npy', sti)

        # WINDOW FUNCTION FILTERING
        size = sti.shape
        # TODO: Use a better 2d convolution function
        sti = self._conv2d(sti, self._filter_win)
        np.save(f'images/stiv/f_{cnt}_1.npy', sti)

        # DETECTION OF PRINCIPAL DIRECTION OF FOURIER SPECTRUM
        sti_ft = np.abs(np.fft.fftshift(np.fft.fft2(sti)))
        # filter vertical and horizontal patterns
        c_x = int(sti_ft.shape[0]/2)
        c_y = int(sti_ft.shape[1]/2)
        sti_ft[c_x - self._vh_filter:c_x + self._vh_filter, :] = 0
        sti_ft[:, c_y - self._vh_filter:c_y + self._vh_filter] = 0
        np.save(f'images/stiv/f_{cnt}_2.npy', sti_ft)
        # transform to polar system
        sti_ft_polar = self._to_polar_system(sti_ft)
        np.save(f'images/stiv/f_{cnt}_3.npy', sti_ft_polar)

        # FILTER IN FREQUENCY DOMAIN
        polar_mask = self._generate_polar_mask(sti_ft_polar)
        sti_ft_polar = sti_ft_polar * polar_mask
        np.save(f'images/stiv/f_{cnt}_4.npy', sti_ft_polar)

        sti_ft_filtered = self._to_polar_system(sti_ft_polar, 'invert')
        np.save(f'images/stiv/f_{cnt}_5.npy', sti_ft_filtered)

        sti_filtered = np.abs(np.fft.ifft2(np.fft.ifftshift(sti_ft_filtered)))
        np.save(f'images/stiv/f_{cnt}_6.npy', sti_filtered)

        sti_filtered = cv2.resize(
                sti_filtered,
                (size[1], size[0]),
                interpolation = cv2.INTER_AREA)

        sti_filtered = np.interp(
                sti_filtered,
                (sti_filtered.min(), sti_filtered.max()),
                (0, 255)
                ).astype(np.uint8)
        return sti_filtered

    def _generate_final_image(self, sti, mask):
        """generate rgb image"""
        new_sti = np.interp(
                sti,
                (sti.min(), sti.max()),
                (0, 255)
                ).astype(np.uint8)
        new_sti = cv2.equalizeHist(sti)
        new_sti = cv2.cvtColor(new_sti, cv2.COLOR_GRAY2RGB)
        new_mask = np.interp(
                mask,
                (mask.min(), mask.max()),
                (0, 255)
                ).astype(np.uint8)

        new_mask = cv2.cvtColor(new_mask, cv2.COLOR_GRAY2RGB)
        out = cv2.add(new_sti, new_mask)

        return out

    @staticmethod
    def _squarify(M):
        (a, b)=M.shape
        if a>b:
            padding=((0,0),(0,a-b))
        else:
            padding=((0,b-a),(0,0))
        return np.pad(M, padding)

    def _calculate_MOT_using_FFT(self, sti):
        """"""
        np.save(f'images/stiv/g_{cnt}_0.npy', sti)
        sti_canny = cv2.Canny(sti, 10, 10)
        np.save(f'images/stiv/g_{cnt}_1.npy', sti_canny)
        sti_padd = self._squarify(sti_canny)
        np.save(f'images/stiv/g_{cnt}_2.npy', sti_padd)
        sti_ft = np.abs(np.fft.fftshift(np.fft.fft2(sti_padd)))
        np.save(f'images/stiv/g_{cnt}_3.npy', sti_ft)
        sti_ft_polar = self._to_polar_system(sti_ft)
        np.save(f'images/stiv/g_{cnt}_4.npy', sti_ft_polar)
        isd = np.sum(sti_ft_polar.T, axis=0)
        freq, _ = self._get_main_freqs(isd)
        angle0 = 2*math.pi *freq / sti_ft_polar.shape[0]
        angle1 = 2*math.pi *freq / sti_ft_polar.shape[1]
        angle = (angle0 + angle1)/2
        velocity = self._get_velocity(angle)
        if self._debug >= 1:
            print("angle:", round(angle, 2))
            print("velocity:", round(velocity, 2))
        mask = np.zeros(sti.shape)
        mask = self._draw_angle(
                mask,
                angle,
                (int(sti.shape[1]/2), int(sti.shape[0]/2)),
                thick=10,
                amplitud=80)

        return velocity, mask

    def _calculate_MOT_using_GMT(self, sti: np.ndarray):
        """
        Calcualte MOT using GMT explained:
        "Development of a non-intrusive and efficient flow monitoring technique:
        The space-time image velocimetry (STIV)"
        """
        window_width = int(self._config['window_shape'][0]/2)
        window_height = int(self._config['window_shape'][1]/2)

        width = sti.shape[0]
        height = sti.shape[1]

        angle_accumulated = 0
        c_total = 0

        # plot vectors
        mask = np.zeros(sti.shape)

        s = window_width
        i = 0
        while s + window_width < width:
            j = 0
            e = window_height
            while e + window_height < height:
                ss = slice(s-window_width, s+window_width)
                ee = slice(e-window_height, e+window_height)
                image_window = sti[ss,ee]
                angle, coherence = self._process_sti(image_window)
                angle_accumulated += (angle * coherence)
                c_total += coherence
                if self._debug >= 2:
                    print((f'- at ({i}, {j}): angle = '
                           f'- in ({s}, {e}): angle = '
                           f'{math.degrees(angle):0.2f}, '
                           f'coherence={coherence:0.2f}, '
                           f'velocity={round(self._get_velocity(angle),2)}'))
                mask = self._draw_angle(mask, angle, (e, s))
                j+=1
                e += int(self._overlap)
            i+=1
            s += int(self._overlap)

        mean_angle = angle_accumulated / c_total

        velocity = self._get_velocity(mean_angle)
        if self._debug  >= 1:
            print("weighted mean angle:", round(math.degrees(mean_angle), 2))
            print("velocity", round(velocity, 2))

        return velocity, mask

    def run(self, show_image=False):
        """Execute"""
        global cnt
        velocities = []
        for idx, sti in enumerate(self._stis):
            if self._debug >= 1:
                print(f'space time image {idx} shape: {sti.shape}')
            cnt = idx
            sti = self._filter_sti(sti)
            # velocity, mask = self._calculate_MOT_using_GMT(sti)
            velocity, mask = self._calculate_MOT_using_FFT(sti)
            velocities.append(velocity)

            final_image = sti + mask
            np.save(f'images/stiv/stiv_final_{idx:02}.npy', final_image)
            cv2.imwrite(
                    f'images/stiv/stiv_final_{idx:02}.png',
                    self._generate_final_image(sti, mask),
                    )
            if show_image:
                cv2.imshow('stiv final', final_image)
                cv2.waitKey(0)
                cv2.destroyAllWindows()

        total = 0
        out_json = {}
        for i, vel in enumerate(velocities):
            total += vel
            out_json[str(i)] = {}
            out_json[str(i)]['velocity'] = vel
        total /= len(velocities)
        if self._debug >= 1:
            print('Total mean velocity:', round(total, 2))
        return out_json


def main(config_path: str, video_identifier: str, show_image=False, debug=0):
    """Execute example of STIV usage."""
    t0 = time.process_time()
    stiv = STIV(config_path, video_identifier, debug)
    t1 = time.process_time()
    ret = stiv.run(show_image)
    t2 = time.process_time()

    print('- STIV\t', t1 - t0)
    print('- stiv.run\t', t2 - t1)
    return ret


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "statio_name",
        help="Name of the station to be analyzed"
    )
    parser.add_argument(
        "video_identifier",
        help="Index of the video of the json config file"
    )
    parser.add_argument(
        '-p',
        '--path',
        help='Path to the config folder',
        type=str,
        default=FOLDER_PATH
    )
    parser.add_argument(
        '-d',
        '--debug',
        help='Activate debug mode',
        type=int,
        default=0
    )
    parser.add_argument(
        '-i',
        '--image',
        action='store_true',
        help='Show every space time image'
    )
    args = parser.parse_args()
    CONFIG_PATH = f'{args.path}/{args.statio_name}.json'
    print(
        main(
            config_path=CONFIG_PATH,
            video_identifier=args.video_identifier,
            show_image=args.image,
            debug=args.debug
        )
    )



"""Configuration."""

from pydantic import BaseModel
import json
from pathlib import Path


class Position(BaseModel):
    """Pixels or meters."""

    x: list[int]
    y: list[int]


class GroundTruth(BaseModel):
    """Ground truth data."""

    position: list[int]
    velocity: float


class ConfigGcp(BaseModel):
    """Configurations GCP."""

    apply: bool
    pixels: Position
    meters: Position
    ground_truth: list[GroundTruth]


class ConfigRoi(BaseModel):
    """Configurations ROI."""

    h1: int
    h2: int
    w1: int
    w2: int


class ConfigImageCorrection(BaseModel):
    """Configuration Image Correction."""

    apply: bool
    k1: float
    c: float
    f: float


class ConfigPreProcessing(BaseModel):
    """Configurations pre-processing."""

    rotate_image: int
    pre_roi: ConfigRoi
    roi: ConfigRoi
    image_correction: ConfigImageCorrection


class ConfigDataset(BaseModel):
    """Configuration dataset."""

    image_dataset: str
    image_number_offset: int
    image_path_prefix: str
    image_path_digits: int
    video_path: str
    width: int
    height: int
    ppm: int
    gcp: ConfigGcp


class ConfigOtvFeatures(BaseModel):
    """Config for OTV Features."""

    maxcorner: int
    qualitylevel: float
    mindistance: int
    blocksize: int


class ConfigOtvLucasKanade(BaseModel):
    """Config for OTV Lucas Kanade."""

    winsize: int
    max_level: int
    max_count: int
    epsilon: float
    flags: int
    radius: int
    min_eigen_threshold: float


class ConfigOtv(BaseModel):
    """Configuration OTV."""

    mask_path: str
    pixel_to_real: float
    partial_min_angle: float
    partial_max_angle: float
    final_min_angle: float
    final_max_angle: float
    final_min_distance: int
    max_features: int
    region_step: int
    resolution: int
    features: ConfigOtvFeatures
    lk: ConfigOtvLucasKanade
    lines: list[int]


class ConfigStivLine(BaseModel):
    """Config for STIV line."""

    start: list[int]
    end: list[int]


class ConfigStiv(BaseModel):
    """Configuration STIV."""

    window_shape: list[int]
    filter_window: int
    overlap: int
    ksize: int
    polar_filter_width: int
    lines: list[ConfigStivLine]


class Config(BaseModel):
    """Config class for awive."""

    dataset: ConfigDataset
    otv: ConfigOtv
    stiv: ConfigStiv
    preprocessing: ConfigPreProcessing

    @staticmethod
    def from_json(file_path: str, video_id: str):
        """Load config from json."""
        return Config(**json.load(Path(file_path).open())[video_id])
